from __future__ import annotations

import contextlib
import logging
import os
import sys
import typing

import httpx
from starlette.responses import Response

import starlette_caches.utils.logging


class ComparableStarletteResponse:
    # As of 0.12, Starlette does not provide a '.__eq__()' implementation
    # for responses yet.

    def __init__(self, response: Response) -> None:
        self.response = response

    def __eq__(self, other: typing.Any) -> bool:
        assert isinstance(other, Response)
        return (
            self.response.body == other.body
            and self.response.raw_headers == other.raw_headers
            and self.response.status_code == other.status_code
        )


class ComparableHTTPXResponse:
    # As of 0.7, HTTPX does not provide a '.__eq__()' implementation
    # for responses yet.

    def __init__(self, response: httpx.Response) -> None:
        self.response = response

    def __eq__(self, other: typing.Any) -> bool:
        assert isinstance(other, httpx.Response)
        return (
            self.response.content == other.content
            and self.response.headers == other.headers
            and self.response.status_code == other.status_code
        )


@contextlib.contextmanager
def override_log_level(log_level: str) -> typing.Iterator[None]:
    os.environ["STARLETTE_CACHES_LOG_LEVEL"] = log_level

    # Force a reload on the logging handlers
    starlette_caches.utils.logging._logger_factory._initialized = False
    starlette_caches.utils.logging.get_logger("starlette_caches")

    try:
        yield
    finally:
        # Reset the logger so we don't have verbose output in all unit tests
        logging.getLogger("starlette_caches").handlers = []


@contextlib.contextmanager
def cleanup_new_imports(prefix: str = "tests.") -> typing.Iterator[None]:
    existing_imports = set(sys.modules)
    try:
        yield
    finally:
        new_imports = set(sys.modules) - existing_imports
        for module in new_imports:
            if module.startswith(prefix):
                del sys.modules[module]



from __future__ import annotations

import math
from typing import TYPE_CHECKING

from starlette.applications import Starlette
from starlette.endpoints import HTTPEndpoint
from starlette.responses import JSONResponse, PlainTextResponse, Response
from starlette.routing import Mount, Route

from starlette_caches.decorators import cache_control
from starlette_caches.middleware import CacheMiddleware

from .resources import cache, special_cache

if TYPE_CHECKING:
    from starlette.requests import Request


async def home(request: Request) -> Response:
    return PlainTextResponse("Hello, world!")


@cache_control(max_age=30, must_revalidate=True)
class Pi(HTTPEndpoint):
    async def get(self, request: Request) -> Response:
        return JSONResponse({"value": math.pi})


class Exp(HTTPEndpoint):
    async def get(self, request: Request) -> Response:
        return JSONResponse({"value": math.e})


async def sub_home(request: Request) -> Response:
    return PlainTextResponse("Hello, sub world!")


sub_app = Starlette(routes=[Route("/", sub_home)])


app = Starlette(
    routes=[
        Route("/", home),
        Route("/pi", CacheMiddleware(Pi, cache=cache)),
        Route("/exp", CacheMiddleware(Exp, cache=special_cache)),
        Mount("/sub", CacheMiddleware(sub_app, cache=cache)),
    ],
)



from __future__ import annotations

import importlib
import math
import typing

import pytest
from starlette.testclient import TestClient

from .utils import cleanup_new_imports

if typing.TYPE_CHECKING:
    from starlette.types import ASGIApp

# TIP: use 'pytest -k <id>' to run tests for a given example application only.
EXAMPLES = [
    pytest.param("tests.examples.functional.starlette", id="starlette"),
]


@pytest.fixture(name="app", params=EXAMPLES)
def fixture_app(request: pytest.FixtureRequest) -> typing.Iterator[ASGIApp]:
    with cleanup_new_imports():
        module: typing.Any = importlib.import_module(request.param)
        yield module.app


@pytest.fixture(name="client")
def fixture_client(app: ASGIApp) -> typing.Iterator[TestClient]:
    with TestClient(app) as client:
        yield client


def test_caching(client: TestClient) -> None:
    r = client.get("/")
    assert r.status_code == 200
    assert r.text == "Hello, world!"
    assert "Expires" not in r.headers
    assert "Cache-Control" not in r.headers
    assert "X-Cache" not in r.headers

    r = client.get("/")
    assert "X-Cache" not in r.headers

    r = client.get("/pi")
    assert r.status_code == 200
    assert r.json() == {"value": math.pi}
    assert r.headers["X-Cache"] == "miss"
    assert "Expires" in r.headers
    assert "Cache-Control" in r.headers
    assert r.headers["Cache-Control"] == "max-age=30, must-revalidate"

    r = client.get("/pi")
    assert r.headers["X-Cache"] == "hit"

    r = client.get("/sub/")
    assert r.status_code == 200
    assert r.text == "Hello, sub world!"
    assert r.headers["X-Cache"] == "miss"
    assert "Expires" in r.headers
    assert "Cache-Control" in r.headers
    assert r.headers["Cache-Control"] == "max-age=120"

    r = client.get("/sub/")
    assert r.headers["X-Cache"] == "hit"

    r = client.get("/exp")
    assert r.status_code == 200
    assert r.json() == {"value": math.e}
    assert r.headers["X-Cache"] == "miss"
    assert "Expires" in r.headers
    assert "Cache-Control" in r.headers
    assert r.headers["Cache-Control"] == "max-age=60"

    r = client.get("/exp")
    assert r.headers["X-Cache"] == "hit"



from aiocache import Cache
from starlette_caches.helpers import CacheHelper
from starlette_caches.middleware import CacheMiddleware
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.requests import Request
from starlette.responses import JSONResponse, Response
from starlette.routing import Route


async def invalidate_cache(request: Request) -> Response:
    cache_helper = CacheHelper(request)
    await cache_helper.invalidate_cache_for("my_endpoint")
    return Response(status_code=204)


async def my_endpoint(request: Request) -> JSONResponse:
    return JSONResponse({"message": "Hello, World!"})


app = Starlette(
    routes=[
        Route("/invalidate", invalidate_cache, methods=["POST"]),
        Route("/", my_endpoint, name="my_endpoint", methods=["GET"]),
    ],
    middleware=[
        Middleware(CacheMiddleware, cache=Cache()),
    ],
)



from __future__ import annotations

import re
import typing
from dataclasses import dataclass

if typing.TYPE_CHECKING:
    from collections.abc import Iterable, Sequence

    from starlette.requests import Request
    from starlette.responses import Response


@dataclass
class Rule:
    """A rule for configuring caching behavior.

    A rule is matched if the request path matches the `match` attribute and the response
    status code matches the `status` attribute.

    If the rule matches, the response will be cached for the duration specified by the
    `ttl` attribute. A value of 0 will disable caching for the response.

    All arguments are optional.
    """

    match: str | re.Pattern | Iterable[str | re.Pattern] = "*"
    """The request path to match.

    If a regular expression is provided, it will be matched against the request path.

    If a sequence is provided, the request path will be matched against each item in the
    sequence.

    If the request path matches any item in the sequence, the rule will match.
    """

    status: int | Iterable[int] | None = None
    """An integer or sequence of integers that match the response status code.

    If the response status code matches any item in the sequence, the rule will match.
    """

    ttl: float | None = None
    """Time-to-live for the cached response in seconds.

    If set, the response will be cached for the specified duration. If not set, the
    default TTL will be used. A value of 0 will disable caching for the response.
    """


def request_matches_rule(
    rule: Rule,
    *,
    request: Request,
) -> bool:
    match = (
        [rule.match] if isinstance(rule.match, (str, re.Pattern)) else list(rule.match)
    )
    for item in match:
        if isinstance(item, re.Pattern):
            if item.match(request.url.path):
                return True
        elif item == "*" or item == request.url.path:
            return True
    return False


def response_matches_rule(rule: Rule, *, request: Request, response: Response) -> bool:
    if not request_matches_rule(rule, request=request):
        return False

    if rule.status is not None:
        statuses = [rule.status] if isinstance(rule.status, int) else rule.status
        if response.status_code not in statuses:
            return False
    return True


def get_rule_matching_request(
    rules: Sequence[Rule], *, request: Request
) -> Rule | None:
    return next(
        (rule for rule in rules if request_matches_rule(rule, request=request)), None
    )


def get_rule_matching_response(
    rules: Sequence[Rule],
    *,
    request: Request,
    response: Response,
) -> Rule | None:
    return next(
        (
            rule
            for rule in rules
            if response_matches_rule(rule, request=request, response=response)
        ),
        None,
    )



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from starlette.applications import Starlette
from starlette.responses import PlainTextResponse, Response
from starlette.routing import Route
from starlette.testclient import TestClient

from starlette_caches.decorators import cache_control

if TYPE_CHECKING:
    from starlette.requests import Request
    from starlette.types import Receive, Scope, Send


def test_cache_control_decorator() -> None:
    @cache_control(stale_if_error=60, must_revalidate=True)
    async def app(scope: Scope, receive: Receive, send: Send) -> None:
        response = PlainTextResponse("Hello, world!")
        await response(scope, receive, send)

    app = Starlette(routes=[Route("/", app)])

    with TestClient(app) as client:
        r = client.get("/")
        assert r.status_code == 200
        assert r.text == "Hello, world!"
        assert r.headers["Cache-Control"] == "stale-if-error=60, must-revalidate"


def test_decorate_starlette_view() -> None:
    with pytest.raises(ValueError, match="does not seem to be an ASGI3 callable"):

        @cache_control(stale_if_error=60)  # type: ignore
        async def home(request: Request) -> Response: ...  # pragma: no cover



from __future__ import annotations

import typing

import pytest

from starlette_caches.utils.misc import is_asgi3

if typing.TYPE_CHECKING:
    from starlette.requests import Request
    from starlette.responses import Response
    from starlette.types import Receive, Scope, Send


class CallableClass:
    async def __call__(
        self, scope: Scope, receive: Receive, send: Send
    ) -> None: ...  # pragma: no cover


callable_instance = CallableClass()


class AwaitableClass:
    def __init__(
        self, scope: Scope, receive: Receive, send: Send
    ) -> None: ...  # pragma: no cover

    def __await__(self) -> None: ...  # pragma: no cover


async def async_function(
    scope: Scope, receive: Receive, send: Send
) -> None: ...  # pragma: no cover


async def view(request: Request) -> Response: ...  # pragma: no cover


@pytest.mark.parametrize(
    ("app", "output"),
    [
        (CallableClass, False),
        (callable_instance, True),
        (AwaitableClass, True),
        (async_function, True),
        (view, False),
        (object(), False),
    ],
)
def test_is_asgi3(*, app: typing.Any, output: bool) -> None:
    assert is_asgi3(app) == output



from __future__ import annotations

import contextlib
import typing

import pytest
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.responses import PlainTextResponse
from starlette.routing import Mount
from starlette.testclient import TestClient

from starlette_caches.middleware import CacheControlMiddleware


@pytest.mark.parametrize(
    ("initial", "kwargs", "result"),
    [
        pytest.param(None, {}, None, id="no-op"),
        pytest.param("stale-if-error=30", {}, "stale-if-error=30", id="copy-initial"),
        pytest.param(None, {"stale_if_error": 60}, "stale-if-error=60", id="add-value"),
        pytest.param(
            "stale-if-error=30",
            {"stale_if_error": 60},
            "stale-if-error=60",
            id="override-value",
        ),
        pytest.param(
            "max-stale=60",
            {"max_stale": False},
            None,
            id="remove-value",
        ),
        pytest.param(None, {"must_revalidate": True}, "must-revalidate", id="add-true"),
        pytest.param(None, {"must_revalidate": False}, None, id="add-false"),
        pytest.param(
            "must-revalidate",
            {"must_revalidate": False},
            None,
            id="remove-false",
        ),
        pytest.param(
            "must-revalidate, max-stale=60, only-if-cached",
            {"stale_if_error": 60, "no_transform": True, "max_stale": False},
            "must-revalidate, only-if-cached, stale-if-error=60, no-transform",
            id="mixed",
        ),
        pytest.param(
            "max-age=60", {"max_age": 30}, "max-age=30", id="override-max-age-1"
        ),
        pytest.param(
            "max-age=30", {"max_age": 60}, "max-age=30", id="override-max-age-2"
        ),
        pytest.param(None, {"public": True}, NotImplementedError),
        pytest.param(None, {"private": True}, NotImplementedError),
    ],
)
def test_cache_control_middleware(
    initial: str | None,
    kwargs: dict,
    result: str | type[BaseException] | None,
) -> None:
    app = Starlette(
        routes=[
            Mount(
                "/",
                PlainTextResponse(
                    "Hello, world!",
                    headers={"Cache-Control": initial} if initial else {},
                ),
            )
        ],
        middleware=[Middleware(CacheControlMiddleware, **kwargs)],
    )
    client = TestClient(app)

    with client:
        if result is NotImplementedError:
            with pytest.raises(NotImplementedError):
                client.get("/")
        else:
            r = client.get("/")
            assert r.status_code == 200
            assert r.text == "Hello, world!"
            if not result:
                assert "Cache-Control" not in r.headers
            else:
                assert r.headers["Cache-Control"] == result


def test_not_http() -> None:
    lifespan_state = None

    @contextlib.asynccontextmanager
    async def lifespan(_: Starlette) -> typing.AsyncIterator[None]:
        nonlocal lifespan_state
        lifespan_state = "started"
        try:
            yield
        finally:
            lifespan_state = "stopped"

    app = Starlette(middleware=[Middleware(CacheControlMiddleware)], lifespan=lifespan)

    with TestClient(app):
        assert lifespan_state == "started"
    assert lifespan_state == "stopped"



import importlib
import typing

import pytest
from starlette.testclient import TestClient
from starlette.types import ASGIApp

from .utils import cleanup_new_imports

EXAMPLES = [
    pytest.param("tests.examples.invalidation.fastapi", id="fastapi"),
    pytest.param("tests.examples.invalidation.starlette", id="starlette"),
]


@pytest.fixture(name="app", params=EXAMPLES)
def app_factory(request: pytest.FixtureRequest) -> typing.Iterator[ASGIApp]:
    with cleanup_new_imports():
        module: typing.Any = importlib.import_module(request.param)
        yield module.app


@pytest.fixture(name="client")
def client_factory(app: ASGIApp) -> typing.Iterator[TestClient]:
    with TestClient(app) as client:
        yield client


def test_response_auto_invalidation(client: TestClient) -> None:
    r = client.get("/")
    assert r.status_code == 200
    assert r.text == "Hello, GET!"
    assert r.headers.get("X-Cache") == "miss"

    r = client.post("/")
    assert r.status_code == 200
    assert r.text == "Hello, POST!"
    assert "X-Cache" not in r.headers

    r = client.get("/")
    assert r.status_code == 200
    assert r.text == "Hello, GET!"
    assert r.headers.get("X-Cache") == "miss"


def test_response_manual_invalidation(client: TestClient) -> None:
    r = client.get("/")
    assert r.status_code == 200
    assert r.text == "Hello, GET!"
    assert r.headers["X-Cache"] == "miss"

    r = client.get("/")
    assert r.headers["X-Cache"] == "hit"

    r = client.post("/invalidate")
    assert r.status_code == 204, r.text

    r = client.get("/")
    assert r.text == "Hello, GET!"
    assert r.headers["X-Cache"] == "miss"


def test_response_invalidation_noop(client: TestClient) -> None:
    r = client.post("/invalidate")
    assert r.status_code == 204, r.text

    r = client.get("/")
    assert r.status_code == 200
    assert r.text == "Hello, GET!"
    assert r.headers["X-Cache"] == "miss"



from starlette.requests import Request
from starlette.responses import Response


class StarletteCachesException(Exception):
    pass


class RequestNotCachable(StarletteCachesException):
    """Raised internally when a request cannot be cached."""

    def __init__(self, request: Request) -> None:
        super().__init__()
        self.request = request


class ResponseNotCachable(StarletteCachesException):
    """Raised internally when a response cannot be cached."""

    def __init__(self, response: Response) -> None:
        super().__init__()
        self.response = response


class DuplicateCaching(StarletteCachesException):
    """Raised when multiple cache middleware were detected."""


class MissingCaching(StarletteCachesException):
    """Raised when no cache middleware was detected."""



"""Miscellaneous utilities and helper functions."""

import base64
import email.utils
import inspect
import typing

TFunc = typing.TypeVar("TFunc", bound=typing.Callable[..., typing.Any])


def http_date(epoch_time: float) -> str:
    """Return a formatted date, for use in HTTP headers.

    See: https://tools.ietf.org/html/rfc7231#section-7.1.1.2
    """
    return email.utils.formatdate(epoch_time, usegmt=True)


def bytes_to_json_string(data: bytes) -> str:
    """Safely convert a bytes object to string for json.

    Given binary data, return a string representation
    that can be safely used in a JSON object.
    """
    # NOTE: we can't just return 'data.decode()', because that won't work
    # if 'data' is not in a given encoding (e.g. utf-8), as is the case
    # when e.g. 'data' is gzip-compressed.
    return base64.encodebytes(data).decode("ascii")


def json_string_to_bytes(value: str) -> bytes:
    """Convert a json string to a bytes object.

    Given a previously-computed JSON-compatible string representation of
    binary data, return the original binary data.
    """
    return base64.decodebytes(value.encode("ascii"))


def has_asgi3_signature(func: typing.Callable) -> bool:
    sig = inspect.signature(func)
    own_parameters = {name for name in sig.parameters if name != "self"}
    return own_parameters == {"scope", "receive", "send"}


def is_asgi3(app: typing.Any) -> bool:
    """Return whether 'app' corresponds to an ASGI3 callable."""
    if inspect.isclass(app):
        constructor = app.__init__  # type: ignore
        return has_asgi3_signature(constructor) and hasattr(app, "__await__")

    if inspect.isfunction(app):
        return inspect.iscoroutinefunction(app) and has_asgi3_signature(app)

    try:
        call = app.__call__
    except AttributeError:
        return False
    else:
        return inspect.iscoroutinefunction(call) and has_asgi3_signature(call)


def kvformat(**kwargs: typing.Any) -> str:
    return " ".join(f"{key}={value}" for key, value in kwargs.items())



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
from aiocache import Cache
from fastapi.routing import Mount
from starlette.applications import Starlette
from starlette.endpoints import HTTPEndpoint
from starlette.responses import PlainTextResponse, Response
from starlette.routing import Route
from starlette.testclient import TestClient

from starlette_caches.decorators import cached
from starlette_caches.middleware import CacheMiddleware

if TYPE_CHECKING:
    from starlette.requests import Request
    from starlette.types import Receive, Scope, Send


def test_decorator_raw_asgi() -> None:
    cache = Cache(ttl=2 * 60)

    @cached(cache=cache)
    async def mount(scope: Scope, receive: Receive, send: Send) -> None:
        response = PlainTextResponse("Hello, world!")
        await response(scope, receive, send)

    app = Starlette(routes=[Mount("/", mount)])

    with TestClient(app) as client:
        r = client.get("/")
        assert r.status_code == 200
        assert r.text == "Hello, world!"
        assert "Expires" in r.headers
        assert "Cache-Control" in r.headers
        assert r.headers["X-Cache"] == "miss"

        r = client.get("/")
        assert r.status_code == 200
        assert r.text == "Hello, world!"
        assert "Expires" in r.headers
        assert "Cache-Control" in r.headers
        assert r.headers["X-Cache"] == "hit"


def test_decorator_starlette_endpoint() -> None:
    cache = Cache(ttl=2 * 60)

    @cached(cache=cache)
    class CachedHome(HTTPEndpoint):
        async def get(self, request: Request) -> Response:
            return PlainTextResponse("Hello, world!")

    class UncachedUsers(HTTPEndpoint):
        async def get(self, request: Request) -> Response:
            return PlainTextResponse("Hello, users!")

    assert isinstance(CachedHome, CacheMiddleware)

    app = Starlette(routes=[Route("/", CachedHome), Route("/users", UncachedUsers)])

    with TestClient(app) as client:
        r = client.get("/")
        assert r.status_code == 200
        assert r.text == "Hello, world!"
        assert "Expires" in r.headers
        assert "Cache-Control" in r.headers
        assert r.headers["X-Cache"] == "miss"

        r = client.get("/")
        assert r.status_code == 200
        assert r.text == "Hello, world!"
        assert "Expires" in r.headers
        assert "Cache-Control" in r.headers
        assert r.headers["X-Cache"] == "hit"

        r = client.get("/users")
        assert r.status_code == 200
        assert r.text == "Hello, users!"
        assert "Expires" not in r.headers
        assert "Cache-Control" not in r.headers
        assert "X-Cache" not in r.headers

        r = client.get("/users")
        assert r.status_code == 200
        assert r.text == "Hello, users!"
        assert "Expires" not in r.headers
        assert "Cache-Control" not in r.headers
        assert "X-Cache" not in r.headers


def test_decorate_starlette_view() -> None:
    cache = Cache(ttl=2 * 60)

    with pytest.raises(ValueError, match="does not seem to be an ASGI3 callable"):

        @cached(cache)  # type: ignore
        async def home(request: Request) -> Response: ...  # pragma: no cover



# ruff: noqa: FA100
import typing
from collections.abc import Mapping

from starlette.datastructures import URL, Headers
from starlette.requests import Request

from .exceptions import MissingCaching
from .middleware import SCOPE_NAME, CacheMiddleware
from .utils.cache import delete_from_cache


class _BaseCacheMiddlewareHelper:
    """Base class for helpers that need access to the `CacheMiddleware` instance."""

    def __init__(self, request: Request) -> None:
        """Initialize the helper with the request and the cache middleware instance.

        Args:
            request: The request object.

        Raises:
            MissingCaching: If the cache middleware instance is not found in the scope
                            or if the cache middleware instance is not an instance of
                            `CacheMiddleware`.

        """
        self.request = request

        if SCOPE_NAME not in request.scope:  # pragma: no cover
            raise MissingCaching(
                "No CacheMiddleware instance found in the ASGI scope. Did you forget "
                "to wrap the ASGI application with `CacheMiddleware`?"
            )

        middleware = request.scope[SCOPE_NAME]
        if not isinstance(middleware, CacheMiddleware):  # pragma: no cover
            raise MissingCaching(
                f"A scope variable named {SCOPE_NAME!r} was found, but it does not "
                "contain a `CacheMiddleware` instance. It is likely that an "
                "incompatible middleware was added to the middleware stack."
            )

        self.middleware = middleware


class CacheHelper(_BaseCacheMiddlewareHelper):
    """Helper class for the `CacheMiddleware`.

    This helper class provides a way to maniuplate the cache middleware.

    If using FastAPI, you can use the `CacheHelper` as a dependency in your endpoint.

    Example:
        ```python
        async def invalidate_cache(helper: Annotated[CacheHelper, Depends()]) -> None:
            await helper.invalidate_cache_for("my_route")
        ```

    """

    async def invalidate_cache_for(
        self,
        url: typing.Union[str, URL],
        *,
        headers: typing.Union[Mapping[str, str], None] = None,
    ) -> None:
        """Invalidate the cache for a given named route or full url.

        `headers` will be used to generate the cache key. The `Vary` header from the
        cached response will determine which headers will be used.

        Args:
            url: The URL to invalidate or name of a starlette route.
            headers: The headers used to generate the cache key.

        """
        if not isinstance(url, URL):
            url = self.request.url_for(url)

        if not isinstance(headers, Headers):
            headers = Headers(headers)

        await delete_from_cache(url, vary=headers, cache=self.middleware.cache)



from typing import Annotated, Any

from aiocache import Cache
from starlette_caches.helpers import CacheHelper
from starlette_caches.middleware import CacheMiddleware
from fastapi import Depends, FastAPI

app = FastAPI()
app.add_middleware(CacheMiddleware, cache=Cache())


@app.post("/invalidate", status_code=204)
async def invalidate_cache(
    cache_helper: Annotated[CacheHelper, Depends()],
) -> None:
    await cache_helper.invalidate_cache_for("my_endpoint")


@app.get("/endpoint")
async def my_endpoint() -> Any:
    return {"message": "Hello, World!"}



from __future__ import annotations

import contextlib
import datetime as dt
import re
import typing
from functools import partial

import pytest
from aiocache import Cache
from fastapi.testclient import TestClient
from starlette.applications import Starlette
from starlette.endpoints import HTTPEndpoint
from starlette.middleware import Middleware
from starlette.middleware.gzip import GZipMiddleware
from starlette.responses import PlainTextResponse, Response, StreamingResponse
from starlette.routing import Route

from starlette_caches.exceptions import DuplicateCaching
from starlette_caches.middleware import CacheMiddleware
from starlette_caches.rules import Rule
from tests.utils import ComparableHTTPXResponse

if typing.TYPE_CHECKING:
    from starlette.requests import Request


async def standard_route(request: Request, *, status_code: int = 200) -> Response:
    return PlainTextResponse("Hello, world!", status_code=status_code)


def test_cache_response() -> None:
    cache = Cache(ttl=2 * 60)

    app = Starlette(
        routes=[Route("/", standard_route)],
        middleware=[Middleware(CacheMiddleware, cache=cache)],
    )

    with TestClient(app) as client:
        r = client.get("/")
        assert r.status_code == 200
        assert r.text == "Hello, world!"
        assert r.headers.pop("X-Cache") == "miss"

        assert "Expires" in r.headers
        expires_fmt = "%a, %d %b %Y %H:%M:%S GMT"
        expires = dt.datetime.strptime(r.headers["Expires"], expires_fmt).replace(
            tzinfo=dt.timezone.utc
        )
        delta: dt.timedelta = expires - dt.datetime.now(tz=dt.timezone.utc)
        assert delta.total_seconds() == pytest.approx(120, rel=1e-2)
        assert "Cache-Control" in r.headers
        assert r.headers["Cache-Control"] == "max-age=120"

        r1 = client.get("/")
        assert r1.headers.pop("X-Cache") == "hit"
        assert ComparableHTTPXResponse(r1) == r

        r2 = client.get("/")
        assert r2.headers.pop("X-Cache") == "hit"
        assert ComparableHTTPXResponse(r2) == r


def test_not_http() -> None:
    lifespan_state = None

    @contextlib.asynccontextmanager
    async def lifespan(_: Starlette) -> typing.AsyncIterator[None]:
        nonlocal lifespan_state
        lifespan_state = "started"
        try:
            yield
        finally:
            lifespan_state = "stopped"

    cache = Cache()
    app = Starlette(
        middleware=[Middleware(CacheMiddleware, cache=cache)],
        lifespan=lifespan,
    )

    with TestClient(app):
        assert lifespan_state == "started"

    assert lifespan_state == "stopped"


def test_non_cachable_request() -> None:
    cache = Cache()
    app = Starlette(
        routes=[Route("/", standard_route, methods=["POST"])],
        middleware=[Middleware(CacheMiddleware, cache=cache)],
    )
    with TestClient(app) as client:
        r = client.post("/")
        assert r.status_code == 200
        assert r.text == "Hello, world!"
        assert "Expires" not in r.headers
        assert "Cache-Control" not in r.headers
        assert "X-Cache" not in r.headers

        r1 = client.post("/")
        assert "X-Cache" not in r1.headers
        assert ComparableHTTPXResponse(r1) == r


@pytest.mark.parametrize(
    ("path", "match_path"),
    [
        ("/cache", "/cache"),
        ("/cache/subpath", re.compile(r"\/cache\/.+")),
    ],
)
def test_cache_match_paths(path: str, match_path: re.Pattern) -> None:
    cache = Cache()
    app = Starlette(
        routes=[
            Route("/", standard_route),
            Route(path, standard_route),
        ],
        middleware=[
            Middleware(
                CacheMiddleware,
                cache=cache,
                rules=[
                    Rule(match_path),
                ],
            )
        ],
    )

    with TestClient(app) as client:
        r = client.get(path)
        assert r.status_code == 200
        assert r.text == "Hello, world!"
        assert r.headers["X-Cache"] == "miss"

        r1 = client.get(path)
        assert r1.status_code == 200
        assert r1.text == "Hello, world!"
        assert r1.headers["X-Cache"] == "hit"

        r2 = client.get("/")
        assert r2.status_code == 200
        assert r2.text == "Hello, world!"
        assert "X-Cache" not in r2.headers


def test_cache_deny_paths() -> None:
    cache = Cache()
    app = Starlette(
        routes=[
            Route("/", standard_route),
            Route("/no_cache", standard_route),
        ],
        middleware=[
            Middleware(
                CacheMiddleware,
                cache=cache,
                rules=[Rule("/no_cache", ttl=0), Rule()],
            )
        ],
    )
    with TestClient(app) as client:
        r = client.get("/no_cache")
        assert r.status_code == 200
        assert r.text == "Hello, world!"
        assert "X-Cache" not in r.headers

        r1 = client.get("/no_cache")
        assert r1.status_code == 200
        assert r1.text == "Hello, world!"
        assert "X-Cache" not in r.headers


def test_use_cached_head_response_on_get() -> None:
    """
    Making a HEAD request should use the cached response for future GET requests.
    """
    cache = Cache()
    app = Starlette(
        routes=[Route("/", standard_route)],
        middleware=[Middleware(CacheMiddleware, cache=cache)],
    )
    with TestClient(app) as client:
        r = client.head("/")
        assert not r.text
        assert r.status_code == 200
        assert "Expires" in r.headers
        assert "Cache-Control" in r.headers
        assert r.headers["X-Cache"] == "miss"

        r1 = client.get("/")
        assert r1.text == "Hello, world!"
        assert r1.status_code == 200
        assert "Expires" in r.headers
        assert "Cache-Control" in r.headers
        assert r1.headers["X-Cache"] == "hit"


def test_rule_exclusion() -> None:
    cache = Cache()
    # 404 status is not included, so it should not be cached.
    rules = [Rule(status=200, ttl=60)]
    app = Starlette(
        routes=[Route("/", partial(standard_route, status_code=404))],
        middleware=[Middleware(CacheMiddleware, cache=cache, rules=rules)],
    )

    with TestClient(app) as client:
        r = client.get("/")
        assert r.status_code == 404
        assert r.text == "Hello, world!"
        assert "X-Cache" not in r.headers

        r1 = client.get("/")
        assert r1.status_code == 404
        assert r1.text == "Hello, world!"
        assert "X-Cache" not in r.headers


def test_rule_stacking() -> None:
    cache = Cache()
    rules = [
        Rule("/", ttl=0),  # don't cache the root path
        Rule(),
    ]
    app = Starlette(
        routes=[
            Route("/", partial(standard_route, status_code=404)),
            Route("/test", partial(standard_route, status_code=404)),
        ],
        middleware=[Middleware(CacheMiddleware, cache=cache, rules=rules)],
    )

    with TestClient(app) as client:
        r = client.get("/")
        assert r.status_code == 404
        assert r.text == "Hello, world!"
        assert "X-Cache" not in r.headers

        r1 = client.get("/")
        assert r1.status_code == 404
        assert r1.text == "Hello, world!"
        assert "X-Cache" not in r.headers

        # /test should be cached
        r = client.get("/test")
        assert r.status_code == 404
        assert r.text == "Hello, world!"
        assert r.headers["X-Cache"] == "miss"

        r1 = client.get("/test")
        assert r1.status_code == 404
        assert r1.text == "Hello, world!"
        assert r1.headers["X-Cache"] == "hit"


@pytest.mark.parametrize(
    "status_code", [201, 202, 307, 308, 400, 401, 403, 500, 502, 503]
)
def test_not_200_ok(status_code: int) -> None:
    """Responses that don't have status code 200 should not be cached."""
    cache = Cache()
    app = Starlette(
        routes=[Route("/", partial(standard_route, status_code=status_code))],
        middleware=[Middleware(CacheMiddleware, cache=cache)],
    )

    with TestClient(app) as client:
        r = client.get("/")
        assert r.status_code == status_code
        assert r.text == "Hello, world!"
        assert "Expires" not in r.headers
        assert "Cache-Control" not in r.headers
        assert "X-Cache" not in r.headers

        r1 = client.get("/")
        assert "X-Cache" not in r1.headers
        assert ComparableHTTPXResponse(r1) == r


def test_streaming_response() -> None:
    """Streaming responses should not be cached."""
    cache = Cache()

    async def body() -> typing.AsyncIterator[str]:
        yield "Hello, "
        yield "world!"

    async def streaming_route(request: Request) -> Response:
        return StreamingResponse(body())

    app = Starlette(
        routes=[Route("/", streaming_route)],
        middleware=[Middleware(CacheMiddleware, cache=cache)],
    )

    with TestClient(app) as client:
        r = client.get("/")
        assert r.status_code == 200
        assert r.text == "Hello, world!"
        assert "X-Cache" not in r.headers

        r = client.get("/")
        assert r.status_code == 200
        assert r.text == "Hello, world!"
        assert "X-Cache" not in r.headers


def test_vary() -> None:
    """
    Sending different values for request headers registered as varying should
    result in different cache entries.
    """
    cache = Cache()

    app = Starlette(
        routes=[Route("/", standard_route)],
        middleware=[
            Middleware(CacheMiddleware, cache=cache),
            Middleware(GZipMiddleware, minimum_size=0),
        ],
    )

    with TestClient(app) as client:
        r = client.get("/", headers={"accept-encoding": "gzip"})
        assert r.headers["X-Cache"] == "miss"
        assert r.status_code == 200
        assert r.text == "Hello, world!"
        assert r.headers.get("vary", "").lower() == "accept-encoding"
        assert r.headers.get("content-encoding") == "gzip"
        assert "Expires" in r.headers
        assert "Cache-Control" in r.headers

        # Different "Accept-Encoding" header => the cached result
        # for "Accept-Encoding: gzip" should not be used.
        r1 = client.get("/", headers={"accept-encoding": "identity"})
        assert r1.headers["X-Cache"] == "miss"
        assert r1.status_code == 200
        assert r1.text == "Hello, world!"
        assert r1.headers.get("vary", "").lower() == "accept-encoding"
        assert "Expires" in r1.headers
        assert "Cache-Control" in r1.headers

        # This "Accept-Encoding" header has already been seen => we should
        # get a cached response.
        r2 = client.get("/", headers={"accept-encoding": "gzip"})
        assert r2.headers["X-Cache"] == "hit"
        assert r2.status_code == 200
        assert r2.text == "Hello, world!"
        assert r.headers.get("vary", "").lower() == "accept-encoding"
        assert r2.headers.get("Content-Encoding") == "gzip"
        assert "Expires" in r2.headers
        assert "Cache-Control" in r2.headers


def test_cookies_in_response_and_cookieless_request() -> None:
    """
    Responses that set cookies shouldn't be cached
    if the request doesn't have cookies.
    """
    cache = Cache()

    async def cookie_route(request: Request) -> Response:
        response = PlainTextResponse("Hello, world!")
        response.set_cookie("session_id", "1234")
        return response

    app = Starlette(
        routes=[Route("/", cookie_route)],
        middleware=[Middleware(CacheMiddleware, cache=cache)],
    )

    with TestClient(app) as client:
        r = client.get("/")
        assert r.status_code == 200
        assert r.text == "Hello, world!"
        # first request is not cached
        assert "X-Cache" not in r.headers

        r = client.get("/")
        assert r.status_code == 200
        assert r.text == "Hello, world!"
        assert r.headers["X-Cache"] == "miss"


def test_duplicate_caching() -> None:
    cache = Cache()
    special_cache = Cache()

    class DuplicateCache(HTTPEndpoint):
        pass

    app = Starlette(
        routes=[
            Route(
                "/duplicate_cache", CacheMiddleware(DuplicateCache, cache=special_cache)
            )
        ],
        middleware=[Middleware(CacheMiddleware, cache=cache)],
    )

    with TestClient(app) as client, pytest.raises(DuplicateCaching):
        client.get("/duplicate_cache")



import pytest
from aiocache import Cache
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.responses import PlainTextResponse
from starlette.routing import Route
from starlette.testclient import TestClient

from starlette_caches.middleware import CacheMiddleware
from tests.utils import override_log_level


def test_logs_debug(capsys: pytest.CaptureFixture) -> None:
    cache = Cache(ttl=2 * 60)
    app = Starlette(
        routes=[Route("/", PlainTextResponse("Hello, world!"))],
        middleware=[Middleware(CacheMiddleware, cache=cache)],
    )

    with TestClient(app) as client, override_log_level("debug"):
        client.get("/")
        client.get("/")

    stderr = capsys.readouterr().err
    miss_line, store_line, hit_line, *_ = stderr.split("\n")
    assert "cache_lookup MISS" in miss_line
    assert "store_in_cache max_age=120" in store_line
    assert "cache_lookup HIT" in hit_line
    assert "get_from_cache request.url='http://testserver/" not in stderr


def test_logs_trace(capsys: pytest.CaptureFixture) -> None:
    cache = Cache(ttl=2 * 60)
    app = Starlette(
        routes=[Route("/", PlainTextResponse("Hello, world!"))],
        middleware=[Middleware(CacheMiddleware, cache=cache)],
    )

    with TestClient(app) as client, override_log_level("trace"):
        client.get("/")

    stderr = capsys.readouterr().err
    assert "cache_lookup MISS" in stderr
    assert "get_from_cache request.url='http://testserver/" in stderr



# noqa: A005
import logging
import os
import sys
import typing

try:  # pragma: no cover
    import click

    # Extra log info for optional coloured terminal outputs.
    HIT_EXTRA = {"color_message": "cache_lookup " + click.style("%s", fg="green")}
    MISS_EXTRA = {"color_message": "cache_lookup " + click.style("%s", fg="yellow")}
except ImportError:  # pragma: no cover
    HIT_EXTRA = {}
    MISS_EXTRA = {}


TRACE_LOG_LEVEL = 5


class Logger(logging.Logger):
    # Stub for type checkers.
    def trace(
        self, message: str, *args: typing.Any, **kwargs: typing.Any
    ) -> None: ...  # pragma: no cover


class LoggerFactory:
    log_level_env_var = "STARLETTE_CACHES_LOG_LEVEL"
    log_line_format = "%(levelname)s [%(asctime)s] %(name)s - %(message)s"
    date_format = "%Y-%m-%d %H:%M:%S"

    def _configure_package_logger(self, log_level: str) -> None:
        assert log_level in ("DEBUG", "TRACE")
        logger = logging.getLogger("starlette_caches")
        logger.setLevel(logging.DEBUG if log_level == "DEBUG" else TRACE_LOG_LEVEL)
        handler = logging.StreamHandler(sys.stderr)
        handler.setFormatter(
            logging.Formatter(fmt=self.log_line_format, datefmt=self.date_format)
        )
        logger.addHandler(handler)

    def get(self, name: str) -> Logger:
        """Get a logger instance, and optionally set up logging."""
        if not getattr(self, "_initialized", False):
            logging.addLevelName(TRACE_LOG_LEVEL, "TRACE")
            log_level = os.environ.get(self.log_level_env_var, "").upper()
            if log_level in ("DEBUG", "TRACE"):
                self._configure_package_logger(log_level)
            self._initialized = True

        logger = logging.getLogger(name)

        def trace(message: str, *args: typing.Any, **kwargs: typing.Any) -> None:
            logger.log(TRACE_LOG_LEVEL, message, *args, **kwargs)

        logger.trace = trace  # type: ignore

        return typing.cast(Logger, logger)


_logger_factory = LoggerFactory()
get_logger = _logger_factory.get



from __future__ import annotations

import typing
from functools import partial

from starlette.datastructures import MutableHeaders
from starlette.requests import Request
from starlette.responses import Response

from .exceptions import DuplicateCaching, RequestNotCachable, ResponseNotCachable
from .rules import Rule
from .utils.cache import (
    INVALIDATING_METHODS,
    CacheDirectives,
    delete_from_cache,
    get_from_cache,
    patch_cache_control,
    store_in_cache,
)
from .utils.logging import HIT_EXTRA, MISS_EXTRA, get_logger
from .utils.misc import kvformat

if typing.TYPE_CHECKING:
    from collections.abc import Sequence

    from aiocache.base import BaseCache as Cache
    from starlette.types import ASGIApp, Message, Receive, Scope, Send


SCOPE_NAME = "__starlette_caches__"

logger = get_logger(__name__)


class CacheMiddleware:
    """Middleware that caches responses.

    This middleware caches responses based on the request path. It can be
    configured with rules that determine which requests and responses should be
    cached. Configure the rules by passing a sequence of `Rule` instances to
    the `rules` argument. See [Rules](usage/rules.md) for more information on how to
    configure rules.

    Args:
        app: The ASGI application to wrap.
        cache: The cache instance to use.
        rules: A sequence of rules for caching behavior.

    """

    def __init__(
        self,
        app: ASGIApp,
        *,
        cache: Cache,
        rules: Sequence[Rule] | None = None,
    ) -> None:
        if rules is None:
            rules = [Rule()]

        self.app = app
        self.cache = cache
        self.rules = rules

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        if SCOPE_NAME in scope:
            raise DuplicateCaching(
                "Another `CacheMiddleware` was detected in the middleware stack.\n"
                "HINT: this exception probably occurred because:\n"
                "- You wrapped an application around `CacheMiddleware` multiple "
                "times.\n"
                "- You tried to apply `@cached()` onto an endpoint, but "
                "the application is already wrapped around a `CacheMiddleware`."
            )

        scope[SCOPE_NAME] = self

        responder = CacheResponder(
            self.app,
            cache=self.cache,
            rules=self.rules,
        )
        await responder(scope, receive, send)


class CacheResponder:
    def __init__(
        self,
        app: ASGIApp,
        *,
        cache: Cache,
        rules: Sequence[Rule],
    ) -> None:
        self.app = app
        self.cache = cache
        self.rules = rules

        self.initial_message: Message = {}
        self.is_response_cachable = True
        self.request: Request | None = None

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        assert scope["type"] == "http"

        self.request = request = Request(scope)

        try:
            response = await get_from_cache(request, cache=self.cache, rules=self.rules)
        except RequestNotCachable:
            if request.method in INVALIDATING_METHODS:
                send = partial(self.send_then_invalidate, send=send)
        else:
            if response is not None:
                logger.debug("cache_lookup %s", "HIT", extra=HIT_EXTRA)
                await response(scope, receive, send)
                return
            send = partial(self.send_with_caching, send=send)
            logger.debug("cache_lookup %s", "MISS", extra=MISS_EXTRA)

        await self.app(scope, receive, send)

    async def send_with_caching(self, message: Message, *, send: Send) -> None:
        if not self.is_response_cachable:
            await send(message)
            return

        if message["type"] == "http.response.start":
            # Defer sending this message until we figured out
            # whether the response can be cached.
            self.initial_message = message
            return

        assert message["type"] == "http.response.body"
        if message.get("more_body", False):
            logger.trace("response_not_cachable reason=is_streaming")
            self.is_response_cachable = False
            await send(self.initial_message)
            await send(message)
            return

        assert self.request is not None
        body = message["body"]
        response = Response(content=body, status_code=self.initial_message["status"])
        # NOTE: be sure not to mutate the original headers directly, as another Response
        # object might be holding a reference to the same list.
        response.raw_headers = list(self.initial_message["headers"])

        try:
            await store_in_cache(
                response, request=self.request, cache=self.cache, rules=self.rules
            )
        except ResponseNotCachable:
            self.is_response_cachable = False
        else:
            # Apply any headers added or modified by 'store_in_cache()'.
            self.initial_message["headers"] = list(response.raw_headers)

        await send(self.initial_message)
        await send(message)

    async def send_then_invalidate(self, message: Message, *, send: Send) -> None:
        # listen for the response start message and invalidate the cache
        # if the request method is POST, PUT, PATCH, DELETE, and if the
        # response status code is 2xx or 3xx
        assert self.request is not None
        if message["type"] == "http.response.start" and 200 <= message["status"] < 400:
            await delete_from_cache(
                self.request.url,
                vary=self.request.headers,
                cache=self.cache,
            )
        await send(message)


class CacheControlMiddleware:
    """Middleware which handles Cache-Control headers for upstream cache proxies.

    Keyword Args:
        max_age (float): The maximum age of the response in seconds.
        public (bool): Not implemented
        private (bool): Not implemented
        **kwargs: Additional Cache-Control directives

    See Also:
        - [Cache-Control - HTTP | MDN](https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Cache-Control)

    """

    def __init__(self, app: ASGIApp, **kwargs: typing.Unpack[CacheDirectives]) -> None:
        self.app = app
        self.kwargs = kwargs

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return
        responder = CacheControlResponder(self.app, **self.kwargs)
        await responder(scope, receive, send)


class CacheControlResponder:
    def __init__(self, app: ASGIApp, **kwargs: typing.Unpack[CacheDirectives]) -> None:
        self.app = app
        self.kwargs = kwargs

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        assert scope["type"] == "http"
        send = partial(self.send_with_caching, send=send)
        await self.app(scope, receive, send)

    async def send_with_caching(self, message: Message, *, send: Send) -> None:
        if message["type"] == "http.response.start":
            logger.trace(f"patch_cache_control {kvformat(**self.kwargs)}")
            headers = MutableHeaders(raw=list(message["headers"]))
            patch_cache_control(headers, **self.kwargs)
            message["headers"] = headers.raw

        await send(message)



from __future__ import annotations

import datetime as dt
import typing

import pytest
import pytest_asyncio
from aiocache import BaseCache, Cache
from starlette.requests import Request
from starlette.responses import PlainTextResponse

from starlette_caches.exceptions import RequestNotCachable, ResponseNotCachable
from starlette_caches.rules import Rule
from starlette_caches.utils.cache import (
    deserialize_response,
    get_cache_key,
    get_from_cache,
    store_in_cache,
)
from tests.utils import ComparableStarletteResponse

if typing.TYPE_CHECKING:
    from starlette.types import Scope

pytestmark = pytest.mark.asyncio


@pytest_asyncio.fixture(name="cache")
async def fixture_cache() -> typing.AsyncIterator[BaseCache]:
    async with Cache() as cache:
        yield cache


@pytest_asyncio.fixture(name="short_cache")
async def fixture_short_cache() -> typing.AsyncIterator[BaseCache]:
    async with Cache(ttl=2 * 60) as cache:
        yield cache


@pytest.mark.parametrize("method", ["GET", "HEAD"])
async def test_get_from_emtpy_cache(cache: BaseCache, method: str) -> None:
    scope: Scope = {
        "type": "http",
        "method": method,
        "path": "/path",
        "headers": [],
    }
    request = Request(scope)
    response = await get_from_cache(request, cache=cache, rules=[Rule()])
    assert response is None


@pytest.mark.parametrize(
    "method", ["POST", "PUT", "PATCH", "DELETE", "OPTIONS", "TRACE"]
)
async def test_non_cachable_methods(cache: BaseCache, method: str) -> None:
    scope: Scope = {
        "type": "http",
        "method": method,
        "path": "/path",
        "headers": [],
    }
    request = Request(scope)
    with pytest.raises(RequestNotCachable):
        await get_from_cache(request, cache=cache, rules=[Rule()])


async def test_store_in_cache(cache: BaseCache) -> None:
    scope: Scope = {
        "type": "http",
        "method": "GET",
        "path": "/path",
        "headers": [],
    }
    request = Request(scope)
    response = PlainTextResponse("Hello, world!")
    assert await get_cache_key(request, method="GET", cache=cache) is None

    await store_in_cache(response, request=request, cache=cache, rules=[Rule()])

    key = await get_cache_key(request, method="GET", cache=cache)
    assert key is not None

    cached_response = deserialize_response(await cache.get(key) or {})

    # We don't care about the "X-Cache" header in the comparison.
    del response.headers["x-cache"]
    del cached_response.headers["x-cache"]

    assert ComparableStarletteResponse(cached_response) == response


@pytest.mark.parametrize(
    "status_code", [201, 202, 307, 308, 400, 401, 403, 500, 502, 503]
)
async def test_non_cachable_status_codes(cache: BaseCache, status_code: int) -> None:
    scope: Scope = {
        "type": "http",
        "method": "GET",
        "path": "/path",
        "headers": [],
    }
    request = Request(scope)
    response = PlainTextResponse("Hello, world!", status_code=status_code)
    with pytest.raises(ResponseNotCachable):
        await store_in_cache(response, request=request, cache=cache, rules=[Rule()])


async def test_non_cachable_zero_ttl(cache: BaseCache) -> None:
    """
    We shouldn't bother caching if the cache TTL is zero.
    """
    cache.ttl = 0
    scope: Scope = {
        "type": "http",
        "method": "GET",
        "path": "/path",
        "headers": [],
    }
    request = Request(scope)
    response = PlainTextResponse("Hello, world!")
    with pytest.raises(ResponseNotCachable):
        await store_in_cache(response, request=request, cache=cache, rules=[Rule()])


async def test_get_from_cache(cache: BaseCache) -> None:
    scope: Scope = {
        "type": "http",
        "method": "GET",
        "path": "/path",
        "headers": [],
    }
    request = Request(scope)
    response = PlainTextResponse("Hello, world!")
    await store_in_cache(response, request=request, cache=cache, rules=[Rule()])

    cached_response = await get_from_cache(request, cache=cache, rules=[Rule()])
    assert cached_response is not None

    # We don't care about the "X-Cache" header in the comparison.
    del response.headers["x-cache"]
    del cached_response.headers["x-cache"]

    assert ComparableStarletteResponse(cached_response) == response
    assert "Expires" in cached_response.headers
    assert "Cache-Control" in cached_response.headers


async def test_default_max_age(cache: BaseCache) -> None:
    scope: Scope = {
        "type": "http",
        "method": "GET",
        "path": "/path",
        "headers": [],
    }
    request = Request(scope)
    response = PlainTextResponse("Hello, world!")
    await store_in_cache(response, request=request, cache=cache, rules=[Rule()])

    cached_response = await get_from_cache(request, cache=cache, rules=[Rule()])
    assert cached_response is not None
    now = dt.datetime.now(tz=dt.timezone.utc)
    http_date_fmt = "%a, %d %b %Y %H:%M:%S GMT"
    expires = dt.datetime.strptime(
        cached_response.headers["Expires"], http_date_fmt
    ).replace(tzinfo=dt.timezone.utc)
    delta: dt.timedelta = expires - now
    one_year = int(dt.timedelta(days=365).total_seconds())
    assert delta.total_seconds() == pytest.approx(one_year)
    assert cached_response.headers["Cache-Control"] == f"max-age={one_year}"


async def test_cache_ttl_max_age(short_cache: BaseCache) -> None:
    scope: Scope = {
        "type": "http",
        "method": "GET",
        "path": "/path",
        "headers": [],
    }
    request = Request(scope)
    response = PlainTextResponse("Hello, world!")
    await store_in_cache(response, request=request, cache=short_cache, rules=[Rule()])

    cached_response = await get_from_cache(request, cache=short_cache, rules=[Rule()])
    assert cached_response is not None
    now = dt.datetime.now(tz=dt.timezone.utc)
    http_date_fmt = "%a, %d %b %Y %H:%M:%S GMT"
    expires = dt.datetime.strptime(
        cached_response.headers["Expires"], http_date_fmt
    ).replace(tzinfo=dt.timezone.utc)
    delta: dt.timedelta = expires - now
    assert delta.total_seconds() == pytest.approx(short_cache.ttl, rel=1e-2)
    ttl = int(short_cache.ttl or 0)
    assert cached_response.headers["Cache-Control"] == f"max-age={ttl}"


async def test_get_from_cache_head(cache: BaseCache) -> None:
    scope: Scope = {
        "type": "http",
        "method": "HEAD",
        "path": "/path",
        "headers": [],
    }
    request = Request(scope)
    response = PlainTextResponse("Hello, world!")
    await store_in_cache(response, request=request, cache=cache, rules=[Rule()])
    assert response.headers["x-cache"] == "miss"

    cached_response = await get_from_cache(request, cache=cache, rules=[Rule()])
    assert cached_response is not None
    assert cached_response.headers["x-cache"] == "hit"

    # We don't care about the "X-Cache" header in the comparison.
    del response.headers["x-cache"]
    del cached_response.headers["x-cache"]

    assert ComparableStarletteResponse(cached_response) == response


async def test_get_from_cache_different_path(cache: BaseCache) -> None:
    scope: Scope = {
        "type": "http",
        "method": "GET",
        "path": "/path",
        "headers": [],
    }
    request = Request(scope)
    response = PlainTextResponse("Hello, world!")
    await store_in_cache(response, request=request, cache=cache, rules=[Rule()])

    other_scope = {**scope, "path": "/other_path"}
    other_request = Request(other_scope)
    cached_response = await get_from_cache(other_request, cache=cache, rules=[Rule()])
    assert cached_response is None


async def test_get_from_cache_vary(cache: BaseCache) -> None:
    scope: Scope = {
        "type": "http",
        "method": "GET",
        "path": "/path",
        "headers": [[b"accept-encoding", b"gzip, deflate"]],
    }
    request = Request(scope)
    # Response indicates that contents of the response at this URL may *vary*
    # depending on the "Accept-Encoding" header sent in the request.
    response = PlainTextResponse("Hello, world!", headers={"Vary": "Accept-Encoding"})
    await store_in_cache(response, request=request, cache=cache, rules=[Rule()])

    # Let's use a different "Accept-Encoding" header,
    # and check that no cached response is found.
    other_scope = {**scope, "headers": [[b"accept-encoding", b"identity"]]}
    other_request = Request(other_scope)
    cached_response = await get_from_cache(other_request, cache=cache, rules=[Rule()])
    assert cached_response is None



from aiocache import Cache
from starlette.applications import Starlette
from starlette.endpoints import HTTPEndpoint
from starlette.middleware import Middleware
from starlette.requests import Request
from starlette.responses import PlainTextResponse, Response
from starlette.routing import Route
from starlette.status import HTTP_204_NO_CONTENT

from starlette_caches.helpers import CacheHelper
from starlette_caches.middleware import CacheMiddleware


class MyRoute(HTTPEndpoint):
    async def get(self, request: Request) -> Response:
        return PlainTextResponse("Hello, GET!")

    async def post(self, request: Request) -> Response:
        return PlainTextResponse("Hello, POST!")


async def invalidation_route(request: Request) -> Response:
    helper = CacheHelper(request)
    # invalidate a named route
    await helper.invalidate_cache_for("my_route")
    return Response(status_code=HTTP_204_NO_CONTENT)


app = Starlette(
    routes=[
        Route("/", MyRoute, name="my_route"),
        Route("/invalidate", invalidation_route, methods=["POST"]),
    ],
    middleware=[
        Middleware(CacheMiddleware, cache=Cache()),
    ],
)



from typing import Annotated

from aiocache import Cache
from fastapi import Depends, FastAPI, status
from fastapi.responses import PlainTextResponse

from starlette_caches.helpers import CacheHelper
from starlette_caches.middleware import CacheMiddleware

app = FastAPI(default_response_class=PlainTextResponse)
app.add_middleware(CacheMiddleware, cache=Cache())


@app.get("/")
async def read_root() -> str:
    return "Hello, GET!"


@app.post("/")
async def create_root() -> str:
    return "Hello, POST!"


@app.post("/invalidate", status_code=status.HTTP_204_NO_CONTENT)
async def invalidate_cache(helper: Annotated[CacheHelper, Depends()]) -> None:
    await helper.invalidate_cache_for("read_root")



"""Utilities that add HTTP-specific to aiocache.

The `store_in_cache()` and `get_from_cache()` helpers are the main pieces of API
defined in this module:

* `store_in_cache()` learns a cache key from a `(request, response)` pair.
* `get_from_cache()` retrieves and uses this cache key for a new `request`.
"""

from __future__ import annotations

import hashlib
import time
import typing
from urllib.request import parse_http_list

from starlette.responses import Response

from ..exceptions import RequestNotCachable, ResponseNotCachable
from ..rules import Rule, get_rule_matching_request, get_rule_matching_response
from .logging import get_logger
from .misc import bytes_to_json_string, http_date, json_string_to_bytes

if typing.TYPE_CHECKING:
    from collections.abc import Sequence

    from aiocache import BaseCache
    from starlette.datastructures import URL, Headers, MutableHeaders
    from starlette.requests import Request

logger = get_logger(__name__)


# https://developer.mozilla.org/en-US/docs/Glossary/Cacheable
CACHABLE_METHODS = frozenset(("GET", "HEAD"))
CACHABLE_STATUS_CODES = frozenset(
    (200, 203, 204, 206, 300, 301, 404, 405, 410, 414, 501)
)
ONE_YEAR = 60 * 60 * 24 * 365

INVALIDATING_METHODS = frozenset(("POST", "PUT", "PATCH", "DELETE"))


class CacheDirectives(typing.TypedDict, total=False):
    max_age: int
    s_maxage: int
    no_cache: bool
    no_store: bool
    no_transform: bool
    must_revalidate: bool
    proxy_revalidate: bool
    must_understand: bool
    private: bool
    public: bool
    immutable: bool
    stale_while_revalidate: int
    stale_if_error: int


async def store_in_cache(
    response: Response,
    *,
    request: Request,
    cache: BaseCache,
    rules: Sequence[Rule],
) -> None:
    """Given a response and a request, store the response in the cache for reuse.

    To do so, a cache key is built from:

    * The absolute URL (including query parameters)
    * Varying headers [^1], as specified specified in the "Vary" header of the response.
      These headers are stored at a key that depends only on the request URL, so that
      they can be retrieved (and checked against) for future requests (without having
      to build and read an uncached response first).

    [^1]: The "Vary" header lists which headers should be taken into account in cache
    systems, because they may result in the server sending in a different response.
    For example, gzip compression requires to add "Accept-Encoding" to "Vary" because
    sending "Accept-Encoding: gzip", and "Accept-Encoding: identity" will result in
    different responses.
    """
    if response.status_code not in CACHABLE_STATUS_CODES:
        logger.trace("response_not_cachable reason=status_code")
        raise ResponseNotCachable(response)

    if not request.cookies and "Set-Cookie" in response.headers:
        logger.trace("response_not_cachable reason=cookies_for_cookieless_request")
        raise ResponseNotCachable(response)

    rule = get_rule_matching_response(rules, request=request, response=response)
    if not rule:
        logger.trace("response_not_cachable reason=rule")
        raise ResponseNotCachable(response)

    ttl = rule.ttl if rule.ttl is not None else cache.ttl

    if ttl == 0:
        logger.trace("response_not_cachable reason=zero_ttl")
        raise ResponseNotCachable(response)

    if ttl is None:
        # From section 14.12 of RFC2616:
        # "HTTP/1.1 servers SHOULD NOT send Expires dates more than
        # one year in the future."
        max_age = ONE_YEAR
        logger.trace(f"max_out_ttl value={max_age!r}")
    else:
        max_age = int(ttl)

    logger.debug(f"store_in_cache max_age={max_age!r}")

    # Store the cached response as a hit. The current request will have it
    # set to miss before returning the response to the client. Future requests
    # will load from the cache, and have it set to hit.
    response.headers["X-Cache"] = "hit"

    cache_headers = get_cache_response_headers(response, max_age=max_age)
    logger.trace(f"patch_response_headers headers={cache_headers!r}")
    response.headers.update(cache_headers)

    cache_key = await learn_cache_key(request, response, cache=cache)
    logger.trace(f"learnt_cache_key cache_key={cache_key!r}")
    serialized_response = serialize_response(response)
    logger.trace(
        f"store_response_in_cache key={cache_key!r} value={serialized_response!r}"
    )
    kwargs = {}
    if ttl is not None:
        kwargs["ttl"] = ttl

    await cache.set(key=cache_key, value=serialized_response, **kwargs)

    # Set X-Cache header to miss for the current request. The next request will
    # load from the cache and have the X-Cache header set to hit.
    response.headers["X-Cache"] = "miss"


async def get_from_cache(
    request: Request, *, cache: BaseCache, rules: Sequence[Rule]
) -> Response | None:
    """Retrieve a cached response based on the cache key associated to the request.

    If no cache key is present yet, or if there is no cached response at
    that key, return `None`.

    A `None` return value indicates that the response for this
    request can (and should) be added to the cache once computed.
    """
    logger.trace(
        f"get_from_cache "
        f"request.url={str(request.url)!r} "
        f"request.method={request.method!r}"
    )
    if request.method not in CACHABLE_METHODS:
        logger.trace("request_not_cachable reason=method")
        raise RequestNotCachable(request)

    rule = get_rule_matching_request(rules, request=request)
    if rule is None:
        logger.trace("request_not_cachable reason=rule")
        raise RequestNotCachable(request)

    logger.trace("lookup_cached_response method='GET'")
    # Try to retrieve the cached GET response (even if this is a HEAD request).
    cache_key = await get_cache_key(request, method="GET", cache=cache)
    if cache_key is None:
        logger.trace("cache_key found=False")
        return None
    logger.trace(f"cache_key found=True cache_key={cache_key!r}")
    serialized_response: dict | None = await cache.get(cache_key)

    # If not present, fallback to look for a cached HEAD response.
    if serialized_response is None:
        logger.trace("lookup_cached_response method='HEAD'")
        cache_key = await get_cache_key(request, method="HEAD", cache=cache)
        assert cache_key is not None
        logger.trace(f"cache_key found=True cache_key={cache_key!r}")
        serialized_response = await cache.get(cache_key)

    if serialized_response is None:
        logger.trace("cached_response found=False")
        return None

    logger.trace(
        f"cached_response found=True key={cache_key!r} value={serialized_response!r}"
    )
    return deserialize_response(serialized_response)


async def delete_from_cache(url: URL, *, vary: Headers, cache: BaseCache) -> None:
    """Clear the cache for the given request."""
    varying_headers_cache_key = generate_varying_headers_cache_key(url, cache=cache)
    varying_headers = await cache.get(varying_headers_cache_key)
    if varying_headers is None:
        # Nothing to do, as there's no cache key associated to this URL.
        return

    for method in "GET", "HEAD":
        cache_key = generate_cache_key(
            url,
            method=method,
            headers=vary,
            varying_headers=varying_headers,
            cache=cache,
        )

        logger.trace(f"clear_cache key={cache_key!r}")
        await cache.delete(cache_key)

    await cache.delete(varying_headers_cache_key)


def serialize_response(response: Response) -> dict:
    """Convert a response to JSON format.

    (This is required as `aiocache` dumps values to JSON before storing them
    in the cache system by default.)
    """
    return {
        "content": bytes_to_json_string(response.body),
        "status_code": response.status_code,
        "headers": dict(response.headers),
    }


def deserialize_response(serialized_response: dict) -> Response:
    """Re-build the original response object from a json-serialized object."""
    return Response(
        content=json_string_to_bytes(serialized_response["content"]),
        status_code=serialized_response["status_code"],
        headers=serialized_response["headers"],
    )


async def learn_cache_key(
    request: Request, response: Response, *, cache: BaseCache
) -> str:
    """Generate a cache key from the requested absolute URL.

    Varying response headers are stored at another key based from the
    requested absolute URL.
    """
    logger.trace(
        "learn_cache_key "
        f"request.method={request.method!r} "
        f"response.headers.Vary={response.headers.get('Vary')!r}"
    )
    url = request.url
    varying_headers_cache_key = generate_varying_headers_cache_key(url, cache=cache)

    cached_vary_headers = set(await cache.get(key=varying_headers_cache_key) or ())
    response_vary_headers = {
        header.lower() for header in parse_http_list(response.headers.get("Vary", ""))
    }

    # workaround for when a route doesn't always add a Vary header
    # Caveat: only effective when a varied requested is sent first
    varying_headers = sorted(response_vary_headers | cached_vary_headers)
    if varying_headers:
        response.headers["Vary"] = ", ".join(varying_headers)

    logger.trace(
        "store_varying_headers "
        f"cache_key={varying_headers_cache_key!r} headers={varying_headers!r}"
    )
    await cache.set(key=varying_headers_cache_key, value=varying_headers)

    return generate_cache_key(
        url,
        method=request.method,
        headers=request.headers,
        varying_headers=varying_headers,
        cache=cache,
    )


async def get_cache_key(request: Request, method: str, cache: BaseCache) -> str | None:
    """Return the cache key where a cached response should be looked up.

    If this request hasn't been served before, return `None` as there definitely
    won't be any matching cached response.
    """
    url = request.url
    logger.trace(f"get_cache_key request.url={str(url)!r} method={method!r}")
    varying_headers_cache_key = generate_varying_headers_cache_key(url, cache=cache)
    varying_headers = await cache.get(varying_headers_cache_key)

    if varying_headers is None:
        logger.trace("varying_headers found=False")
        return None
    logger.trace(f"varying_headers found=True headers={varying_headers!r}")

    return generate_cache_key(
        request.url,
        method=method,
        headers=request.headers,
        varying_headers=varying_headers,
        cache=cache,
    )


def generate_cache_key(
    url: URL,
    method: str,
    headers: Headers,
    varying_headers: list[str],
    cache: BaseCache,
) -> str:
    """Generate a cache key from the request full URL and varying response headers.

    Note that the given `method` may be different from that of the request, e.g.
    because we're trying to find a response cached from a previous GET request
    while this one is a HEAD request. (This is OK because web servers will strip content
    from responses to a HEAD request before sending them on the wire.)
    """
    assert method in CACHABLE_METHODS

    ctx = hashlib.md5(usedforsecurity=False)
    for header in varying_headers:
        value = headers.get(header)
        if value is not None:
            ctx.update(value.encode())
    vary_hash = ctx.hexdigest()

    url_hash = hashlib.md5(str(url).encode("ascii"), usedforsecurity=False).hexdigest()

    return cache.build_key(f"cache_page.{method}.{url_hash}.{vary_hash}")


def generate_varying_headers_cache_key(url: URL, cache: BaseCache) -> str:
    """Generate a cache key from the requested absolute URL.

    Suitable for associating varying headers to a requested URL.
    """
    url_hash = hashlib.md5(str(url.path).encode("ascii"), usedforsecurity=False)
    return cache.build_key(f"varying_headers.{url_hash.hexdigest()}")


def get_cache_response_headers(response: Response, *, max_age: int) -> dict[str, str]:
    """Return caching-related headers to add to a response."""
    assert max_age >= 0, "Can't have a negative cache max-age"
    headers = {}

    if "Expires" not in response.headers:
        headers["Expires"] = http_date(time.time() + max_age)

    patch_cache_control(response.headers, max_age=max_age)

    return headers


def patch_cache_control(
    headers: MutableHeaders, **kwargs: typing.Unpack[CacheDirectives]
) -> None:
    """Patch headers with an extended version of the initial Cache-Control header.

    Appends all keyword arguments to the Cache-Control header.

    True values are added as flags, while false values are omitted.
    """
    cache_control: dict[str, typing.Any] = {}
    value: typing.Any
    for field in parse_http_list(headers.get("Cache-Control", "")):
        try:
            key, value = field.split("=")
        except ValueError:  # noqa: PERF203
            cache_control[field] = True
        else:
            cache_control[key] = value

    if "max-age" in cache_control and "max_age" in kwargs:
        kwargs["max_age"] = min(int(cache_control["max-age"]), kwargs["max_age"])

    if "public" in kwargs:
        raise NotImplementedError(
            "The 'public' cache control directive isn't supported yet."
        )

    if "private" in kwargs:
        raise NotImplementedError(
            "The 'private' cache control directive isn't supported yet."
        )

    for key, value in kwargs.items():
        key = key.replace("_", "-")
        cache_control[key] = value

    directives: list[str] = []
    for key, value in cache_control.items():
        if value is False:
            continue
        if value is True:
            directives.append(key)
        else:
            directives.append(f"{key}={value}")

    patched_cache_control = ", ".join(directives)

    if patched_cache_control:
        headers["Cache-Control"] = patched_cache_control
    else:
        del headers["Cache-Control"]



from __future__ import annotations

import re

import pytest
from starlette.datastructures import URL
from starlette.requests import Request
from starlette.responses import Response

from starlette_caches.rules import (
    Rule,
    get_rule_matching_request,
    get_rule_matching_response,
    request_matches_rule,
    response_matches_rule,
)


def mock_request(path: str) -> Request:
    return Request(
        scope={
            "type": "http",
            "method": "GET",
            "path": path,
            "headers": {},
            "query_string": b"",
            "client": ("127.0.0.1", 12345),
            "server": ("127.0.0.1", 8000),
            "scheme": "http",
            "root_path": "",
            "app": None,
            "url": URL(path),
        }
    )


@pytest.mark.parametrize("match", ["/test", re.compile(r"^/test")])
def test_request_matches_rule(match: str | re.Pattern) -> None:
    rule = Rule(match=match)
    request = mock_request("/test")
    assert request_matches_rule(rule, request=request)


def test_request_matches_rule_with_wildcard() -> None:
    rule = Rule(match="*")
    request = mock_request("/any")
    assert request_matches_rule(rule, request=request)


def test_request_does_not_match_rule() -> None:
    rule = Rule(match="/test")
    request = mock_request("/other")
    assert not request_matches_rule(rule, request=request)


def test_response_matches_rule_with_status() -> None:
    rule = Rule(match="/test", status=200)
    request = mock_request("/test")
    response = Response(status_code=200)
    assert response_matches_rule(rule, request=request, response=response)


def test_response_does_not_match_rule_with_status() -> None:
    rule = Rule(match="/test", status=404)
    request = mock_request("/test")
    response = Response(status_code=200)
    assert not response_matches_rule(rule, request=request, response=response)


def test_get_rule_matching_request() -> None:
    rules = [Rule(match="/test1"), Rule(match="/test2")]
    request = mock_request("/test2")
    rule = get_rule_matching_request(rules, request=request)
    assert rule is not None
    assert rule.match == "/test2"


def test_get_rule_matching_response() -> None:
    rules = [Rule(match="/test1", status=200), Rule(match="/test2", status=404)]
    request = mock_request("/test2")
    response = Response(status_code=404)
    rule = get_rule_matching_response(rules, request=request, response=response)
    assert rule is not None
    assert rule.match == "/test2"
    assert rule.status == 404



from __future__ import annotations

import functools
import sys
import typing

from .middleware import CacheControlMiddleware, CacheMiddleware
from .utils.misc import is_asgi3

if sys.version_info >= (3, 10):  # pragma: no cover
    from typing import ParamSpec
else:  # pragma: no cover
    from typing_extensions import ParamSpec

if typing.TYPE_CHECKING:
    from starlette.types import ASGIApp


_P = ParamSpec("_P")


class _MiddlewareFactory(typing.Protocol[_P]):
    def __call__(
        self, app: ASGIApp, *args: _P.args, **kwargs: _P.kwargs
    ) -> ASGIApp: ...


def _middleware_to_decorator(
    cls: _MiddlewareFactory[_P],
) -> typing.Callable[_P, typing.Callable[[ASGIApp], ASGIApp]]:
    def decorator(
        *args: _P.args, **kwargs: _P.kwargs
    ) -> typing.Callable[[ASGIApp], ASGIApp]:
        def wrap(app: ASGIApp) -> ASGIApp:
            _validate_asgi3(app)
            middleware = cls(app, *args, **kwargs)
            return _wrap_in_middleware(app, middleware)

        return wrap

    return decorator


def _wrap_in_middleware(app: ASGIApp, middleware: ASGIApp) -> ASGIApp:
    # Use `updated=()` to prevent copying `__dict__` onto `middleware`.
    # (If `app` is a middleware itself and has a `.app` attribute, it would be copied
    # onto `middleware`, effectively removing `app` from the middleware chain.)
    return functools.wraps(app, updated=())(middleware)


def _validate_asgi3(app: ASGIApp) -> None:
    if not is_asgi3(app):
        raise ValueError(
            f"{app!r} does not seem to be an ASGI3 callable. "
            "Did you try to apply this decorator to a framework-specific view "
            "function? (It can only be applied to ASGI callables.)"
        )


cached = _middleware_to_decorator(CacheMiddleware)
"""Wrap an ASGI endpoint with [starlette_caches.middleware.CacheMiddleware][].

This decorator provides the same behavior as `CacheMiddleware`,
but at an endpoint level.

Raises 'ValueError' if the wrapped callable isn't an ASGI application.
"""


cache_control = _middleware_to_decorator(CacheControlMiddleware)
"""Wrap an ASGI endpoint with [starlette_caches.middleware.CacheControlMiddleware][].

This decorator provides the same behavior as `CacheControlMiddleware`,
but at an endpoint level.

Raises 'ValueError' if the wrapped callable isn't an ASGI application.
"""
